#!/usr/bin/env python

"""
Authentication/authorisation helper classes and functions.

Copyright (C) 2004, 2005, 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

--------

The authentication token produced in this module typically employs the username
and a secret. If other things are added in addition to the username, it should
not be possible to combine them in a way which causes "collisions" between
distinct username-plus-extra-data inputs.
"""

import base64
import md5
import hmac
try:
    from hashlib import sha1, sha256
except ImportError:
    sha256 = None
    try:
        from sha import new as sha1
    except ImportError:
        sha1 = None

class UserInfo:

    """
    A class used to represent user information in terms of the authentication
    scheme employed and the user details.
    """

    def __init__(self, auth_header):

        """
        Initialise the object with the value of the 'auth_header' - that is, the
        HTTP Authorization header.
        """

        self.scheme, auth_details = auth_header.split(" ")
        if self.scheme == "Basic":

            # NOTE: Assume that no username or password contains ":".

            self.username, self.password = base64.decodestring(auth_details).split(":")

        else:

            # NOTE: Other schemes not yet supported.

            self.username, self.password = None, None

# Classes providing support for authentication resources.

class Authenticator:

    """
    A simple authenticator with no other purpose than to return the status of an
    authentication request.
    """

    def __init__(self, credentials):

        """
        Initialise the authenticator with a registry of 'credentials'.

        The 'credentials' must be an object which supports tests of the form
        '(username, password) in credentials'.
        """

        self.credentials = credentials

    def authenticate(self, trans, username, password):

        """
        Authenticate the sender of the transaction 'trans', returning a true
        value if they are recognised, or a false value otherwise. Use the
        'username' and 'password' supplied as credentials.
        """

        # Check against the class's credentials.

        return (username, password) in self.credentials

class Verifier:

    """
    An authenticator which can only verify an authentication attempt with
    existing credentials, not check new credentials.
    """

    def __init__(self, secret_key, cookie_name=None):

        """
        Initialise the authenticator with a 'secret_key' and an optional
        'cookie_name'.
        """

        self.secret_key = secret_key
        self.cookie_name = cookie_name or "LoginAuthenticator"

    def _encode(self, username):
        return username.replace(":", "%3A")

    def _decode(self, encoded_username):
        return encoded_username.replace("%3A", ":")

    def authenticate(self, trans):

        """
        Authenticate the originator of 'trans', returning a true value if
        successful, or a false value otherwise.
        """

        # Test the token from the cookie against a recreated token using the
        # given information.

        details = self.get_username_and_token(trans)
        if details is None:
            return 0
        else:
            username, token = details
            return token == get_token(self._encode(username), self.secret_key)

    def get_username_and_token(self, trans):

        "Return the username and token for the user."

        cookie = trans.get_cookie(self.cookie_name)
        if cookie is None or cookie.value is None:
            return None
        else:
            return self._decode(cookie.value.split(":")[0]), cookie.value

    def set_token(self, trans, username):

        "Set an authentication token in 'trans' with the given 'username'."

        trans.set_cookie_value(
            self.cookie_name,
            get_token(self._encode(username), self.secret_key),
            path="/"
        )

    def unset_token(self, trans):

        "Unset the authentication token in 'trans'."

        trans.delete_cookie(self.cookie_name)

class LoginAuthenticator(Authenticator, Verifier):

    """
    An authenticator which sets authentication tokens.
    """

    def __init__(self, secret_key, credentials, cookie_name=None):

        """
        Initialise the authenticator with a 'secret_key', the authenticator's registry of
        'credentials' and an optional 'cookie_name'.

        The 'credentials' must be an object which supports tests of the form
        '(username, password) in credentials'.
        """

        Authenticator.__init__(self, credentials)
        Verifier.__init__(self, secret_key, cookie_name)

    def authenticate(self, trans, username, password):

        """
        Authenticate the sender of the transaction 'trans', returning a true
        value if they are recognised, or a false value otherwise. Use the
        'username' and 'password' supplied as credentials.
        """

        valid = Authenticator.authenticate(self, trans, username, password)
        if valid:
            self.set_token(trans, username)
        return valid

def get_token(plaintext, secret_key):

    """
    Return a string containing an authentication token made from the given
    'plaintext' and 'secret_key'.
    """

    # NOTE: Using "safe" encoding to deal with Unicode plaintext.

    return plaintext + ":" + md5.md5(plaintext.encode("iso-8859-1") + secret_key).hexdigest()

# OpenID token verification.
# NOTE: Add SHA256 usage for associations.

if sha1 is not None:

    def get_openid_token(items, secret_key):

        """
        Return a string containing the 'items' encoded using the given
        'secret_key'.
        """

        plaintext = "\n".join([(key + ":" + value) for (key, value) in items]) + "\n"

        # NOTE: Using "safe" encoding to deal with Unicode plaintext.

        hash = hmac.new(secret_key, plaintext.encode("iso-8859-1"), sha1)
        return base64.standard_b64encode(hash.digest())

    def check_openid_signature(fields, secret_key):

        """
        Return whether information in the given 'fields' (a mapping from names
        to lists of values) is signed using the 'secret_key'.
        """

        signed_names = fields["openid.signed"][0].split(",")
        return fields["openid.sig"][0] == make_openid_signature(signed_names, fields, secret_key)

    def make_openid_signature(signed_names, fields, secret_key):

        """
        Make and return a signature using the 'signed_names' to indicate which
        values from the 'fields' (a mapping from names to lists of values) shall
        be signed using the 'secret_key'.
        """

        items = []
        for name in signed_names:
            items.append((name, fields["openid." + name][0]))
        return get_openid_token(items, secret_key)

# vim: tabstop=4 expandtab shiftwidth=4
